/*
This walk does the following:

1) Build Up Code Table
2) callTable is a hash table with keys=nodes and values=linked list
   where the links list is the orders of arguments in a call
   Each value in the linked list is a string of the expression
3) ccTable is used in the if-else while statements to keep track
   of where in the compound statement there are
4) plugTable is similar to ccTable and functionTable ... this is meant for show and exit
5) inputTable is similar to ccTable and functionTable ... this is meant to show receive
6) variableTable is similar to ccTable and functionTable ... but is meant for variable*
7) tupleTable
8) pushcallTable ... linked the function push to its statement,
    so that its printed just before the statement. It contains
    a linked list of strings
9) in codeTable when I have a string that begins with a "@" it means
    that its tuple assignmenet.
    I get rid of the "@" when I print to file 
*/
package wig.code;

import wig.*;
import wig.node.*;
import wig.analysis.*;
import wig.symbol.*;
import wig.compiler.*;
import java.util.*;
import java.io.*;

public class CodeWalk extends DepthFirstAdapter
{

    private SymbolTable GlobalTable;
    private Hashtable cstToScopeMap;
    private Hashtable typeTable;
    private Hashtable codeTable;
    private Hashtable functionTable;
    private Hashtable callTable;
    private Hashtable ccTable;
    private Hashtable plugTable;
    private Hashtable inputTable;
    private Hashtable variableTable;
    private Hashtable tupleTable;
    private Hashtable pushcallTable;
    private boolean htmlconst_started;
    private boolean functionReturned;
    private int ccCounter;
    private Node currentDocument;
    private Node currentVariable;
    private Node currentTuple;
    private AAssignExp currentAssign;
    private PrintWriter out;

    public void print(String x)
    {
        out.print(x);
    }
    public void println(String x)
    {
        out.println(x);
    }

    public CodeWalk(AnalysisDataManager manager, PrintWriter out)
    {
        GlobalTable = manager.getGlobalTable();
        cstToScopeMap = manager.getCstToScopeMap();
        typeTable = manager.getTypeTable();
        codeTable = manager.getCodeTable();
        functionTable = manager.getFunctionTable();
        callTable = manager.getCallTable();
        htmlconst_started = false;
        ccTable = manager.getCCTable();
        plugTable = manager.getPlugTable();
        inputTable = manager.getInputTable();
        variableTable = manager.getVariableTable();
        tupleTable = manager.getTupleTable();
        pushcallTable = manager.getPushCallTable();
        this.out=out;
    }

    /**********************************************
    * finds the closest scope to a node
    **********************************************/
    private SymbolTable getSymbolTable(Node node)
    {
        while(!cstToScopeMap.containsKey(node))
        {
            node = node.parent();
            if(node==null)
                break;
        }
        if(node==null)
        {
            return GlobalTable;
        }
        return (SymbolTable) cstToScopeMap.get(node);
    }

    /**********************************************
    * get the fields of a schema hold
    **********************************************/
    private String[] getFieldsOf(String name)
    {
        //get the Node defining the schema
        Node node = GlobalTable.seekSymbol(name).getNode();
        //get the Symbol Table that define the field inside the schema
        Hashtable schmea_table = ((SymbolTable) cstToScopeMap.get(node)).getTable();

        String[] fieldName = new String[schmea_table.size()];

        Enumeration schmea_enum = schmea_table.elements();
        int i=0;
        while(schmea_enum.hasMoreElements())
        {
            fieldName[i] = ((Symbol)schmea_enum.nextElement()).getIdentifier();
            i++;
        }

        return fieldName;
    }


    public void inAService(AService node)
    {
        // print the top.php to the output file
        // top.php contains the common functions for all the php files we generate
        try
        {
            BufferedReader top = new BufferedReader(
                new FileReader(getClass().getResource("top.php").getPath()));
            String wholeFile="",line;
            while((line=top.readLine())!=null)
            {
                wholeFile += line + "\n";
            }
            top.close();
            print(wholeFile);
        }
        catch(IOException ioE)
        {
            PrintError.error("top.php File Error",0);
        }

        println("\n\n<?");

        //state all globa variables right after the schema
        outGlobalVariables();
    }

    /* define all the global variables after schemas*/
    public void outGlobalVariables()
    {
        Enumeration iterator = GlobalTable.getTable().elements();
        println("\n/*Declare Global Variables*/");
        Symbol s;
        String str ="";
        // the global variables only will define when start the session for first time
        println("if($started!=1){");
        while(iterator.hasMoreElements())
        {
            s = (Symbol) iterator.nextElement();
            // skip the symbol if it is not variable
            if(!s.getKind().equals(Const.variable))
                continue;
            if(!s.getType().equals(Const.tInt)&&!s.getType().equals(Const.tBool)&&
                    !s.getType().equals(Const.tVoid)&&!s.getType().equals(Const.tString))
            {
                /*it must be a schema type*/
                str = "  $globals[\""+s.getIdentifier()+"\"] = new schema_";
                str += s.getType().getType()+"();";
                println(str);
            }
            else
            {
                println("  $globals[\""+s.getIdentifier()+"\"];");
            }
        }
        println("}\n");
    }

    public void outAService(AService node)
    {
        println("?>\n\n");
        
        /* Write the control flow for the session and function
         * We did the control flow like assembly language
         * It put the session and function on the stack and return the value in retstack
         * the two stacks are store in the session.
         * we also have a cc flag for every instruction. It is a visited flag.
         * where cc set to true, it will not be run when a session or function resume.
         * We did that because there don't have goto statment php.
         */
        println("<?");
        println("/*code to manage starting the function on top of the stack*/");
        println("if(count($stack)==0){");
        println("  if(class_exists(\"session_\".$firstsession)){");
        println("    $code = \"array_push(\\$stack, new session_$firstsession());\";");
        println("    eval($code);");
        println("    $stack[count($stack)-1]->run();");
        println("    while(count($stack)>0){");
        println("      $stack[count($stack)-1]->run();");
        println("    }");
        println("  } else {");
        println("    ?><html><body>");
        println("    NO SESSION SELECTED!");
        println("    <p>Please Select the following Session:<br>");

        // print out the available sessions link for user if no session are specified
        Enumeration iterator = GlobalTable.getTable().elements();
        Symbol s;
        String str ="";
        while(iterator.hasMoreElements())
        {
            s = (Symbol) iterator.nextElement();
            if(s.getKind().equals(Const.session))
                println("    <a href=\"<?=$PHP_SELF?>?"+s.getIdentifier()
                    +"\"><?=$PHP_SELF?>?"+s.getIdentifier()+"<br>");
        }

        println("    </body></html><?");
        println("  }");
        println("} else {");
        println("  while(count($stack)>0){");
        println("    $stack[count($stack)-1]->run();");
        println("  }");
        println("}");
        println("?>");
    }

    /*output the htmml consts*/
    public void outAHtml(AHtml node)
    {
        ListIterator list = node.getHtmlbody().listIterator();
        ListIterator attrs,htmlBody;
        PHtmlbody member;
        ATagStartHtmlbody opentag;
        ATagEndHtmlbody endtag;
        AHoleHtmlbody hole;
        AInputHtmlbody input;
        String select;
        
        /* 
         * In Print out of Html, we put the html body to a output buffer and copy it 
         * back to a variable. We make use the "?>" & "<?" feature in php.
         */
        if(!htmlconst_started)
        {
            println("ob_start();");
            htmlconst_started = true;
        }
        println("?>");

        /* 
         * print out the PHP code for the Html Const
         * basically, print out the whole string in the html body.
         * but some space is dropped
         */
        while(list.hasNext())
        {
            member = (PHtmlbody)list.next();
            if(member instanceof ATagStartHtmlbody)
            {
                opentag = (ATagStartHtmlbody) member;
                print(opentag.getLt().getText());
                print(opentag.getIdentifier().getText()+" ");
                attrs = opentag.getAttribute().listIterator();
                while(attrs.hasNext())
                {
                    String string=attrs.next().toString();
                    print(string);
                }
                print(opentag.getGt().getText());
            }
            else if(member instanceof ATagEndHtmlbody)
            {
                endtag = (ATagEndHtmlbody) member;
                print(endtag.getLtSlash().getText());
                print(endtag.getIdentifier().getText());
                print(endtag.getGt().getText());
            }
            else if(member instanceof AHoleHtmlbody)
            {
                hole = (AHoleHtmlbody) member;
                print(hole.getLtBracket().getText());
                print(hole.getIdentifier().getText());
                print(hole.getGtBracket().getText());
            }
            else if(member instanceof AInputHtmlbody)
            {
                input = (AInputHtmlbody) member;
                print(input.getLt().getText());
                print(input.getInput().getText()+" ");
                attrs = input.getAttribute().listIterator();
                while(attrs.hasNext())
                {
                    print((String) codeTable.get(attrs.next()));
                }
                print(input.getGt().getText());
            }
            else if(member instanceof ASelectHtmlbody)
            {
                select = member.toString();
                select = select.replaceAll("< ","<").replaceAll("<\\/ ","</");
                select = select.replaceAll(" >",">");
                print(select);
            }
            else
            {
                print(member.toString());
            }
        }

        println("");
        println("<?");
        println("$htmlconst[\""+node.getIdentifier().getText()+"\"] = ob_get_contents();");
        println("ob_clean();");
    }

    public void outAAttrAttribute(AAttrAttribute node)
    {
        codeTable.put(node,node.getAttr().toString());
    }

    /* 
     * from outANameAttribute to outAAssignAttribute
     * Copy everything in the node to the codeTable
     */
    public void outANameAttribute(ANameAttribute node)
    {
        String str = "";
        str += node.getName().getText();
        str += node.getAssign().getText();
        str += node.getAttr().toString();
        codeTable.put(node,str);
    }

    public void outATypeAttribute(ATypeAttribute node)
    {
        String str ="";
        str += node.getType().getText();
        str += node.getAssign().getText();
        str +=node.getAttr().toString();
        codeTable.put(node,str);
    }

    public void outAAssignAttribute(AAssignAttribute node)
    {
        String str ="";
        str += node.getLeftAttr().toString();
        str += node.getAssign().getText();
        str+= node.getRightAttr().toString();
        codeTable.put(node,str);
    }


    /*
     * out put the schemas
     */
    public void outASchema(ASchema node)
    {
        /*
         * we build up a class for every schema
         */
        Enumeration iterator = ((SymbolTable)cstToScopeMap.get(node)).getTable().elements();
        println("");
        println("class schema_"+node.getIdentifier().getText()+"{");
        Symbol s;
        while(iterator.hasMoreElements())
        {
            s = (Symbol)iterator.nextElement();
            println("  var $"+s.getIdentifier()+";");
        }
        println("  //class constructor");
        println("  function schema_"+node.getIdentifier().getText()+"(){}");
        println("}\n");
    }


    /* out functions*/
    public void inAFunction(AFunction node)
    {
        /*
         * We build up a class for each function, so that it can be put into stack
         * in the class, it contains a run method. it implement the origianl function.
         * We did the similar thing in session
         */
        println("");
        println("class func_"+node.getIdentifier().getText()+" {");
        println("  //$cc");
        println("  //$local");
        println("  //$name");
        println("  //function class constructor");
        println("  function func_"+node.getIdentifier().getText()+"($argsarray){");
        println("    if(!isset($this->cc)) $this->cc = array();");
        println("    if(!isset($this->local)) $this->local = array();");

        // input arguments is compacted in a array and send to the function.
        // we get it back in here
        Iterator args = ((LinkedList)functionTable.get(node)).listIterator();
        int i=0;
        String I="";
        while(args.hasNext())
        {
            I = (new Integer(i)).toString();
            println("    $this->local[\""+(String)args.next()+"\"] = $argsarray["+I+"];");
            i++;
        }

        println("  }");
        println("  //function running code");
        println("  function run(){ ");
        println("    global $globals;");
        println("    global $_POST;");
        println("    global $htmlconst;");
        println("    global $exit;");
        println("    global $stack;");
        ccCounter = 0;
        functionReturned = false;
    }

    public void outAFunction(AFunction node)
    {
        println((String)codeTable.get(node.getCompoundstm()));
        // Return is implemented by setResult(inside top.php).
        // it push the return value to the retstack.
        if(!functionReturned)
            println("      setResult(\"\"); return;");
        println("  }");
        println("}");
    }


    /* out functions*/
    public void inASession(ASession node)
    {
        //same as a function
        println("");
        println("class session_"+node.getIdentifier().getText()+" {");
        println("  //$cc");
        println("  //$local");
        println("  //$name");
        println("  function session_"+node.getIdentifier().getText()+"(){");
        println("    if(!isset($this->cc)) $this->cc = array();");
        println("    if(!isset($this->local)) $this->local = array();");
        println("  }");
        println("  function run(){ ");
        println("    global $globals;");
        println("    global $_POST;");
        println("    global $htmlconst;");
        println("    global $exit;");
        println("    global $stack;");
        ccCounter = 0;
    }

    public void outASession(ASession node)
    {
        println((String)codeTable.get(node.getCompoundstm()));
        println("  array_pop($stack);\n");
        println("  }");
        println("}");
    }


    /**** lvalue *****/

    /* convert identifier to it corresponding php name */
    public void inASimpleLvalue(ASimpleLvalue node)
    {
        String str = "";
        if(GlobalTable.defsSymbol(node.getIdentifier().getText()))
        { /*global var*/
            str = "$globals[\""+node.getIdentifier().getText();
            str += "\"]";
            codeTable.put(node,str);
        }
        else
        { /*local var*/
            str = "$this->local[\""+node.getIdentifier().getText();
            str += "\"]";
            codeTable.put(node,str);
        }
    }


    /* convert identifier to it corresponding php name */
    public void inAQualifiedLvalue(AQualifiedLvalue node)
    {
        String str = "";
        if(GlobalTable.defsSymbol(node.getLeft().getText()))
        {
            /*global var*/
            str = "$globals[\""+node.getLeft().getText();
            str += "\"]->"+node.getRight().getText();
            codeTable.put(node,str);
        }
        else
        {
            /*local var*/
            str = "$this->local[\""+node.getLeft().getText();
            str += "\"]->"+node.getRight().getText();
            codeTable.put(node,str);
        }
    }

    /*** base_exp****/
    public void outAParenBaseExp(AParenBaseExp node)
    {
        String str = "";
        str += "(";
        str += (String)  codeTable.get(node.getExp());
        str += ")";
        codeTable.put(node,str);
    }

    public void outAStringBaseExp(AStringBaseExp node)
    {
        codeTable.put(node,node.getStringconst().getText());
    }
    public void outAFalseBaseExp(AFalseBaseExp node)
    {
        codeTable.put(node,"false");
    }
    public void outATrueBaseExp(ATrueBaseExp node)
    {
        codeTable.put(node,"true");
    }
    public void outAIntBaseExp(AIntBaseExp node)
    {
        codeTable.put(node,node.getIntconst().getText());
    }

    public void outALvalueBaseExp(ALvalueBaseExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getLvalue()));
    }


    public void inACallBaseExp(ACallBaseExp node)
    {
        callTable.put(node,new LinkedList());
    }

    public void outACallBaseExp(ACallBaseExp node)
    {
        String str = "";
        str =  "  if(!$this->cc["+(new Integer(ccCounter)).toString()+"]){\n";
        str += "    $argsarray = array();\n";
        
        // compact the variables to a array
        Iterator iter = ((LinkedList)callTable.get(node)).listIterator();
        while(iter.hasNext())
        {
            str += "    $argsarray[] = ";
            str += (String) iter.next();
            str += ";\n";
        }
        
        //push the function object to the stack
        str += "    array_push($stack, new func_";
        str += node.getIdentifier().getText();
        str += "($argsarray));\n";
        str += "    $this->cc["+(new Integer(ccCounter)).toString();
        str += "] = true;\n";
        str += "    return;\n}\n";
        /*
         * return to the control flow in the bottom part of the php file.
         * it will call the function run and pop itself from the stack.
         * (like the OS switch out the preivous process and switch in another process)
         */
        
        ccCounter++;


        /*manage pushcallTable*/
        Node pnode = node;
        while(!(pnode instanceof PStm))
        {
            pnode = pnode.parent();
        }
        /*
         * for each statement that have a function call, it will contains a call list
         * so that the call inside the statement will be called in sequence
         * eg. result= callA(callB()+callC());
         * callB() will be run first then callC and then callA
         */
        if(!pushcallTable.containsKey(pnode))
        {
            pushcallTable.put(pnode,new LinkedList());
        }
        
        LinkedList list = (LinkedList) pushcallTable.get(pnode);
        list.add(str);

        /*get the result of the result stack*/
        codeTable.put(node,"getResult()");
    }

    public void inATupleBaseExp(ATupleBaseExp node)
    {
        currentTuple = node;
        tupleTable.put(node,new LinkedList());
    }

    public void outATupleBaseExp(ATupleBaseExp node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "\n  if(!$this->cc["+cc.toString()+"]){\n";
        if(currentAssign==null)
        {
            PrintError.error("Code Generation(1) = should have caught this in type check",
                node.getTuple().getLine());
        }
        
        // currentAssign is set in the inAAssignExp function (this file line 1024)
        String var = (String) codeTable.get(currentAssign.getLvalue());
        //tupleTable contain all the field and its value (see outAFieldvalue)
        Iterator iter = ((LinkedList)tupleTable.get(node)).listIterator();
        while(iter.hasNext())
        {
            str += "      "+var+"->";
            str += (String)iter.next();
        }
        str+= "    $this->cc["+cc.toString()+"]=true;\n}";
        str = "@" + str; /*mark the string that its a tuple */
        codeTable.put(node,str);
    }

    /**** field value *****/
    public void outAFieldvalue(AFieldvalue node)
    {
        // put the field and the value to tupleTable and pass to outATupleBaseExp(
        LinkedList list = (LinkedList) tupleTable.get(currentTuple);
        String str = node.getIdentifier().getText()+" = ";
        str += (String) codeTable.get(node.getExp());
        str += ";\n";
        list.add(str);
    }

    /**** exps ****/
    /* created an ordered list for arguments in a call*/
    public void outAOneExps(AOneExps node)
    {
        Node pnode = node;
        while(!(pnode instanceof ACallBaseExp))
        {
            pnode = pnode.parent();
        }
        ACallBaseExp currentCall = (ACallBaseExp) pnode;
        LinkedList list = (LinkedList)callTable.get(currentCall);
        list.add((String)codeTable.get(node.getExp()));
    }

    public void outAManyExps(AManyExps node)
    {
        Node pnode = node;
        while(!(pnode instanceof ACallBaseExp))
        {
            pnode = pnode.parent();
        }
        ACallBaseExp currentCall = (ACallBaseExp) pnode;
        LinkedList list = (LinkedList)callTable.get(currentCall);
        list.add((String)codeTable.get(node.getExp()));
    }

    /**** unary_exp ****/
    public void outADefaultUnaryExp(ADefaultUnaryExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getBaseExp()));
    }

    public void outANegUnaryExp(ANegUnaryExp node)
    {
        String str = "";
        str += "(-1*";
        str += (String) codeTable.get(node.getBaseExp());
        str += ")";
        codeTable.put(node,str);
    }

    public void outANotUnaryExp(ANotUnaryExp node)
    {
        String str = "";
        str += "(!";
        str += (String) codeTable.get(node.getBaseExp());
        str += ")";
        codeTable.put(node,str);
    }

    /**** tuple_exp ***/
    public void outADefaultTupleExp(ADefaultTupleExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getUnaryExp()));
    }

    public void outAKeepTupleExp(AKeepTupleExp node)
    {
        String temp;
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "\n  if(!$this->cc["+cc.toString()+"]){\n";
        if(currentAssign==null)
        {
            PrintError.error("Code Generation(1) = should have caught this in type check",
                node.getKeep().getLine());
        }

        // Get the tuple in in the left and right hand side of "=" in string
        String var = (String)codeTable.get(currentAssign.getLvalue());
        String var2= (String)codeTable.get(node.getTupleExp());

        /*
         * Get all the fields of the left hand side tuple
         * In Type checking, we have make sure that the right hand side 
         * and left hand side tuple's field is equal in type and number.
         */
        Type lvalueType = (Type) typeTable.get(currentAssign.getLvalue());
        String[] fields = getFieldsOf(lvalueType.getType());

        for(int i=0; i<fields.length ;i++)
        {
            str += "      "+var+"->"+fields[i];
            str += "="+var2+"->"+fields[i]+";\n";
        }
        str+= "    $this->cc["+cc.toString()+"]=true;\n    }";
        str = "@" + str +"@"; /*mark the string that its a tuple*/
        codeTable.put(node,str);
    }

    public void outARemoveTupleExp(ARemoveTupleExp node)
    {
        String temp;
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "\n  if(!$this->cc["+cc.toString()+"]){\n";
        if(currentAssign==null)
        {
            PrintError.error("Code Generation(1) = should have caught this in type check",
                node.getRemove().getLine());
        }

        // same a the outAKeepTupleExp
        String var = (String)codeTable.get(currentAssign.getLvalue());
        String var2= (String)codeTable.get(node.getTupleExp());

        Type lvalueType = (Type) typeTable.get(currentAssign.getLvalue());
        String[] fields = getFieldsOf(lvalueType.getType());

        for(int i=0; i<fields.length ;i++)
        {
            str += "      "+var+"->"+fields[i];
            str += "="+var2+"->"+fields[i]+";\n";
        }
        str+= "    $this->cc["+cc.toString()+"]=true;\n    }";
        str = "@" + str +"@"; /*mark the string that its a tuple*/
        codeTable.put(node,str);
    }

    public void outAKeepManyTupleExp(AKeepManyTupleExp node)
    {
        String temp;
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "\n  if(!$this->cc["+cc.toString()+"]){\n";
        if(currentAssign==null)
        {
            PrintError.error("Code Generation(1) = should have caught this in type check",
                node.getKeep().getLine());
        }

        // same a the outAKeepTupleExp
        String var = (String)codeTable.get(currentAssign.getLvalue());
        String var2= (String)codeTable.get(node.getTupleExp());

        Type lvalueType = (Type) typeTable.get(currentAssign.getLvalue());
        String[] fields = getFieldsOf(lvalueType.getType());

        for(int i=0; i<fields.length ;i++)
        {
            str += "      "+var+"->"+fields[i];
            str += "="+var2+"->"+fields[i]+";\n";
        }
        str+= "    $this->cc["+cc.toString()+"]=true;\n    }";
        str = "@" + str +"@"; /*mark the string that its a tuple*/
        codeTable.put(node,str);
    }

    public void outARemoveManyTupleExp(ARemoveManyTupleExp node)
    {
        String temp;
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "\n  if(!$this->cc["+cc.toString()+"]){\n";
        if(currentAssign==null)
        {
            PrintError.error("Code Generation(1) = should have caught this in type check",
                node.getRemove().getLine());
        }

        // same a the outAKeepTupleExp
        String var = (String)codeTable.get(currentAssign.getLvalue());
        String var2= (String)codeTable.get(node.getTupleExp());

        Type lvalueType = (Type) typeTable.get(currentAssign.getLvalue());
        String[] fields = getFieldsOf(lvalueType.getType());

        for(int i=0; i<fields.length ;i++)
        {
            str += "      "+var+"->"+fields[i];
            str += "="+var2+"->"+fields[i]+";\n";
        }
        str+= "    $this->cc["+cc.toString()+"]=true;\n    }";
        str = "@" + str +"@"; /*mark the string that its a tuple*/
        codeTable.put(node,str);
    }

    /**** join_exp ****/
    public void outADefaultJoinExp(ADefaultJoinExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getTupleExp()));
    }

    public void outAJoinJoinExp(AJoinJoinExp node)
    {
        String temp;
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str="";
        String str2="";
        if(currentAssign==null)
        {
            PrintError.error("Code Generation(1) = should have caught this in type check",
                node.getJoin().getLine());
        }

        PTupleExp left = node.getLeft();
        PJoinExp right = node.getRight();
        // Get the three tuple code in string
        String var = (String)codeTable.get(currentAssign.getLvalue());
        String varLeft= (String)codeTable.get(left);
        String varRight= (String)codeTable.get(right);

        // get the Type of the tuple, by that get the contains of all the fields in schema
        Type leftType = (Type) typeTable.get(left);
        String[] leftfields = getFieldsOf(leftType.getType());
        String[] rightfields = new String[0];
        //Check for multiple join
        if(!(right instanceof AJoinJoinExp))
        {
            // no multiple join
            Type rightType = (Type) typeTable.get(right);
            rightfields = getFieldsOf(rightType.getType());
        }
        else
        {
            //get the code and drop the "@"
            str2=((String)codeTable.get(right)).replaceAll("(^.*@)|(@.*$)","");
        }

        for(int i=0; i<leftfields.length ;i++)
        {
            str += "      "+var+"->"+leftfields[i];
            str += "="+varLeft+"->"+leftfields[i]+";\n";
        }
        str +="\n";
        for(int i=0; i<rightfields.length ;i++)
        {
            str += "      "+var+"->"+rightfields[i];
            str += "="+varRight+"->"+rightfields[i]+";\n";
        }
        //str+= "    $this->cc["+cc.toString()+"]=true;\n    }";
        str = "@" + str + str2 +"@"; /*mark the string that its a tuple*/
        str = str.replaceFirst(";\n@$","@"); //get off the ";"
        codeTable.put(node,str);
    }

    /***** mult_exp ****/
    public void outADefaultMultExp(ADefaultMultExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getJoinExp()));
    }

    public void outAModMultExp(AModMultExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "%";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outAMultMultExp(AMultMultExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "*";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outADivMultExp(ADivMultExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "/";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }



    /***** add_exp ****/
    public void outADefaultAddExp(ADefaultAddExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getMultExp()));
    }


    public void outAPlusAddExp(APlusAddExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        PMultExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        String str = "";
        String op = "";
        if(left_type.equals(right_type)&&left_type.equals(Const.tString))
        {
            op = "."; // "add" two string
        }
        else
        {
            op = "+";
        }
        str += (String) codeTable.get(node.getLeft());
        str += op;
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outAMinusAddExp(AMinusAddExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "-";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    /***** cmp_exp ****/
    public void outADefaultCmpExp(ADefaultCmpExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getAddExp()));
    }


    public void outAEqCmpExp(AEqCmpExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "==";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outANeqCmpExp(ANeqCmpExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "!=";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outALtCmpExp(ALtCmpExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "<";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outAGtCmpExp(AGtCmpExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += ">";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outALteqCmpExp(ALteqCmpExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "<=";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    public void outAGteqCmpExp(AGteqCmpExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += ">=";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }


    /***** and_exp ****/
    public void outADefaultAndExp(ADefaultAndExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getCmpExp()));
    }

    public void outAAndAndExp(AAndAndExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "&&";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }

    /***** or_exp ****/
    public void outADefaultOrExp(ADefaultOrExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getAndExp()));
    }

    public void outAOrOrExp(AOrOrExp node)
    {
        String str = "";
        str += (String) codeTable.get(node.getLeft());
        str += "||";
        str += (String) codeTable.get(node.getRight());
        codeTable.put(node,str);
    }


    /***** exp ****/
    public void outADefaultExp(ADefaultExp node)
    {
        codeTable.put(node,(String)codeTable.get(node.getOrExp()));
    }

    public void inAAssignExp(AAssignExp node)
    {
        currentAssign = node;
    }

    public void outAAssignExp(AAssignExp node)
    {
        String str = "";
        String orexp;
        orexp = (String) codeTable.get(node.getOrExp());
        orexp = orexp.trim();
        //check for tuple expression
        if(orexp.indexOf("@")==-1)
        {
            //normal experssion (add "=" and orexp in front of orexp)
            str += (String) codeTable.get(node.getLvalue());
            str += "=";
            str += orexp;
        }
        else
        {
            //tuple experssion (just drop the @)
            str = orexp.replaceAll("(^.*@)|(@.*$)","");
        }
        codeTable.put(node,str);
        currentAssign=null;
    }



    /***** stm  and stm_no_short_if *****/
    public void outAExpStm(AExpStm node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;

        //in every statment, it has a cc flag (visited or not)
        String str = "   if(!$this->cc["+cc.toString()+"]){ ";
        str += (String)codeTable.get(node.getExp());
        str += "; $this->cc["+cc.toString()+"]=true;}\n";

        codeTable.put(node,str);
    }

    public void outAExpStmNoShortIf(AExpStmNoShortIf node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;

        String str = "    if(!$this->cc["+cc.toString()+"]){ ";
        str += (String)codeTable.get(node.getExp());
        str += "; $this->cc["+cc.toString()+"]=true;}\n";

        codeTable.put(node,str);
    }

    public void inAIfStm(AIfStm node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        ccCounter++;
        ccTable.put(node,cc);
    }

    public void outAIfStm(AIfStm node)
    {
        Integer cc = (Integer) ccTable.get(node);
        Integer cc1 = new Integer(cc.intValue()+1);

        String str = "";
        // if this "if statment" haven't been run
        str += "    if(!$this->cc["+cc.toString()+"]){\n";
        // cc1: "if statment" is entered(but exited by function call or show document)
        // go into it without checking the condition
        str += "      if($this->cc["+cc1.toString()+"]||(";
        str += (String)codeTable.get(node.getExp());
        str += ")){\n";
        //set the cc1 to true, means this branch is entered
        str += "        $this->cc["+cc1.toString()+"]=true; \n";
        str += (String)codeTable.get(node.getStm());
        str += "    }\n";
        str += "  } $this->cc["+cc.toString()+"]=true;\n";
        codeTable.put(node,str);

    }

    public void inAIfelseStm(AIfelseStm node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        ccCounter++;
        ccTable.put(node,cc);
    }

    public void outAIfelseStm(AIfelseStm node)
    {
        Integer cc = (Integer) ccTable.get(node);
        Integer cc1 = new Integer(cc.intValue()+1);
        Integer last = new Integer(ccCounter-1);
        String str = "";
        // similar to if statment
        str += " if(!$this->cc["+cc.toString()+"]&&($this->cc["+cc1.toString()+"]||(";
        str += (String)codeTable.get(node.getExp());
        str += "))){\n";
        str += "        $this->cc["+cc1.toString()+"]=true; \n";
        str += (String) codeTable.get(node.getThenStm());
        str += "    } else {     \n$this->cc["+cc.toString()+"]=true; \n";
        str += (String) codeTable.get(node.getElseStm());
        // set all the statement inside the two branch to visited to make sure the untaken
        // branch will not been run
        str += "  } for($p="+cc.toString()+";$p<="+last.toString()+";$p++) $this->cc[$p] = true; \n";
        codeTable.put(node,str);
    }


    public void inAIfelseStmNoShortIf(AIfelseStmNoShortIf node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        ccCounter++;
        ccTable.put(node,cc);
    }

    public void outAIfelseStmNoShortIf(AIfelseStmNoShortIf node)
    {
        Integer cc = (Integer) ccTable.get(node);
        Integer cc1 = new Integer(cc.intValue()+1);
        Integer last = new Integer(ccCounter-1);
        String str = "";
        // similar to if statment
        str += " if(!$this->cc["+cc.toString()+"]&&($this->cc["+cc1.toString()+"]||(";
        str += (String)codeTable.get(node.getExp());
        str += "))){\n";
        str += "        $this->cc["+cc1.toString()+"]=true; \n";
        str += (String) codeTable.get(node.getThenStm());
        str += "    } else {     \n$this->cc["+cc.toString()+"]=true; \n";
        str += (String) codeTable.get(node.getElseStm());
        str += "  } for($p="+cc.toString()+";$p<="+last.toString()+";$p++) $this->cc[$p] = true; \n";
        codeTable.put(node,str);
    }

    public void outACompStm(ACompStm node)
    {
        codeTable.put(node,(String) codeTable.get(node.getCompoundstm()));
    }

    public void outACompStmNoShortIf(ACompStmNoShortIf node)
    {
        codeTable.put(node,(String) codeTable.get(node.getCompoundstm()));
    }

    public void inAWhileStm(AWhileStm node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        ccCounter++;
        ccTable.put(node,cc);
    }

    public void outAWhileStm(AWhileStm node)
    {
        Integer cc = (Integer) ccTable.get(node);
        int cc2 = cc.intValue()+1;
        Integer ccstart = new Integer(cc.intValue());
        Integer ccend = new Integer(ccCounter-1);
        String str = "";
        // similar to if statment
        str += "  while(!$this->cc["+cc+"]&&($this->cc["+cc2+"] || (";
        str += (String) codeTable.get(node.getExp());
        str += "))){\n";
        str += "    $this->cc["+cc2+"]=true;\n";
        str += (String) codeTable.get(node.getStm());
        // before get out from the while loop it set all the statement to unvisited
        str += "\n   for($k="+ccstart.toString()+";";
        str += "$k<="+ccend.toString()+";$k++){ $this->cc[$k] = false; }\n";
        /*
         * This return is for calling the function in the condition experssion
         * 
         * when there are function call in the condition expression
         * outACallBaseExp will be call and build up the call function outside the while 
         * loop. However, it will put a getResult() method inside the condition expression.
         *
         * in the end of the while loop, it will have a for loop to set all the statement 
         * inside the while loop to unvisited. It will set the function call statement to
         * unvisited too. And we just need to return the control to the control and it will
         * call the function in each iteration of the while loop.
         */
        str += "    return;\n";
        str += "  }";
        str += "  $this->cc["+cc.toString()+"]=true;\n  ";
        codeTable.put(node,str);
    }

    public void inAWhileStmNoShortIf(AWhileStmNoShortIf node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        ccTable.put(node,cc);
    }

    public void outAWhileStmNoShortIf(AWhileStmNoShortIf node)
    {
        Integer cc = (Integer) ccTable.get(node);
        Integer ccstart = new Integer(cc.intValue());
        Integer ccend = new Integer(ccCounter-1);
        String str = "";
        // see outAWhileStm
        str += "  while($this->cc["+cc.toString()+"]||(";
        str += (String) codeTable.get(node.getExp());
        str += ")){\n";
        str += "    $this->cc["+cc.toString()+"]=true;\n";
        str += (String) codeTable.get(node.getStmNoShortIf());
        str += "\n   for($k="+ccstart.toString()+";";
        str += "$k<="+ccend.toString()+";$k++){ ";
        str += "$this->cc[$k] = false; }\n";
        str += "}";
        str += "  $this->cc["+cc.toString()+"]=true;\n";

        codeTable.put(node,str);

    }

    public void outAReturnStm(AReturnStm node)
    {
        functionReturned = true;
        String str = "      setResult(\"\"); return;\n";
        codeTable.put(node,str);
    }

    public void outAReturnStmNoShortIf(AReturnStmNoShortIf node)
    {
        functionReturned = true;
        String str = "      setResult(\"\"); return;\n";
        codeTable.put(node,str);
    }

    public void outARetexpStm(ARetexpStm node)
    {
        functionReturned = true;
        String str = "      setResult(";
        str += (String) codeTable.get(node.getExp());
        str += "); return;\n";
        codeTable.put(node,str);
    }

    public void outARetexpStmNoShortIf(ARetexpStmNoShortIf node)
    {
        functionReturned = true;
        String str = "      setResult(";
        str += (String) codeTable.get(node.getExp());
        str += "); return;\n";
        codeTable.put(node,str);
    }


    /****plug,exit and receive in stm and stm_no_short_if ****/
    // Get a inputTable and plugTable
    public void inAShowStm(AShowStm node)
    {
        currentDocument = node;
        plugTable.put(node,new LinkedList());
        inputTable.put(node,new LinkedList());
    }

    public void inAExitStm(AExitStm node)
    {
        currentDocument = node;
        plugTable.put(node,new LinkedList());
        inputTable.put(node,new LinkedList());
    }

    public void inAShowStmNoShortIf(AShowStmNoShortIf node)
    {
        currentDocument = node;
        plugTable.put(node,new LinkedList());
        inputTable.put(node,new LinkedList());
    }

    public void inAExitStmNoShortIf(AExitStmNoShortIf node)
    {
        currentDocument = node;
        plugTable.put(node,new LinkedList());
        inputTable.put(node,new LinkedList());
    }

    public void outAPlug(APlug node)
    {
        // generate a plug array for the plugs
        LinkedList list = (LinkedList)plugTable.get(currentDocument);
        String str = "    $plugarray[\""+node.getIdentifier().getText()+"\"] = ";
        str += (String) codeTable.get(node.getExp());
        str += ";\n";
        list.add(str);
    }

    public void outAIdDocument(AIdDocument node)
    {
        LinkedList list = (LinkedList)plugTable.get(currentDocument);
        String str = "    plugHtml($plugarray,$htmlconst[\"";
        str += node.getIdentifier().getText()+"\"]);\n";
        list.add(str);
    }

    public void outAPlugDocument(APlugDocument node)
    {
        LinkedList list = (LinkedList)plugTable.get(currentDocument);
        String str = "    plugHtml($plugarray,$htmlconst[\"";
        str += node.getIdentifier().getText()+"\"]);\n";
        list.add(str);
    }


    public void outAInput(AInput node)
    {
        // get the input from the global variable $_POST in php
        LinkedList list = (LinkedList) inputTable.get(currentDocument);
        String str = "    "+((String)codeTable.get(node.getLvalue()));
        str += " = $_POST[\""+node.getIdentifier().getText()+"\"];\n";
        list.add(str);
    }

    public void outAShowStm(AShowStm node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "";
        str += "   if(!$this->cc["+cc.toString()+"]){\n";
        str += "     $this->cc["+cc.toString()+"] = true;\n";
        
        //plug the output
        Iterator iter = ((LinkedList)plugTable.get(node)).listIterator();
        str += "    $plugarray = array();\n ";

        while(iter.hasNext())
        {
            str += (String)iter.next();
        }
        str += "   }\n";

        // Get the input
        cc = new Integer(ccCounter);
        ccCounter++;
        str += "   if(!$this->cc["+cc.toString()+"]){\n";
        str += "     $this->cc["+cc.toString()+"] = true;\n";
        iter = ((LinkedList)inputTable.get(node)).listIterator();
        while(iter.hasNext())
        {
            str += (String)iter.next();
        }
        str += "   }\n";

        codeTable.put(node,str);

    }

    public void outAShowStmNoShortIf(AShowStmNoShortIf node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "";
        str += "   if(!$this->cc["+cc.toString()+"]){\n";
        str += "     $this->cc["+cc.toString()+"] = true;\n";
                
        //plug the output
        Iterator iter = ((LinkedList)plugTable.get(node)).listIterator();
        str += "    $plugarray = array();\n ";
        while(iter.hasNext())
        {
            str += (String)iter.next();
        }
        str += "   }\n";

        // Get the input
        cc = new Integer(ccCounter);
        ccCounter++;
        str += "   if(!$this->cc["+cc.toString()+"]){\n";
        str += "     $this->cc["+cc.toString()+"] = true;\n";
        iter = ((LinkedList)inputTable.get(node)).listIterator();
        while(iter.hasNext())
        {
            str += (String)iter.next();
        }
        str += "   }\n";

        codeTable.put(node,str);

    }


    public void outAExitStm(AExitStm node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "";
        //similar to outAShowStm but no input and need to set the exit flag
        str += "    if(!$this->cc["+cc.toString()+"]){\n";
        str += "      $exit = true;\n";
        str += "      $this->cc["+cc.toString()+"] = true;\n";
        str += "      $plugarray = array();\n ";
        Iterator iter = ((LinkedList)plugTable.get(node)).listIterator();
        while(iter.hasNext())
        {
            str += (String)iter.next();
        }
        str += "    }\n";
        codeTable.put(node,str);
    }

    public void outAExitStmNoShortIf(AExitStmNoShortIf node)
    {

        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "";
        str += "    if(!$this->cc["+cc.toString()+"]){\n";
        str += "      $exit = true;\n";
        str += "      $this->cc["+cc.toString()+"] = true;\n";
        str += "      $plugarray = array();\n ";
        Iterator iter = ((LinkedList)plugTable.get(node)).listIterator();
        while(iter.hasNext())
        {
            str += (String)iter.next();
        }
        str += "    }\n";
        codeTable.put(node,str);

    }

    /**** variable ****/

    public void inAOneIdentifiers(AOneIdentifiers node)
    {

        Node tmpnode=node;
        /*keep going until you hit a variable;*/
        while(tmpnode instanceof PIdentifiers)
        {
            tmpnode = tmpnode.parent();
        }

        /*get rid of case keepmany and removemany from tuples*/
        if(!(tmpnode instanceof AVariable))
        {
            return;
        }
        
        // Global variables have been defined in outGlobalVariables (line 152)

        SymbolTable symtable = getSymbolTable(node);
        Symbol s = symtable.getSymbol(node.getIdentifier().getText());
        String str = "";
        if(!s.getType().equals(Const.tInt)&&!s.getType().equals(Const.tBool)&&
                !s.getType().equals(Const.tVoid)&&!s.getType().equals(Const.tString))
        {
            /*its a tuple type*/
            LinkedList list = (LinkedList) variableTable.get(currentVariable);
            str = "      $this->local[\""+node.getIdentifier().getText()+"\"] = new schema_";
            str += s.getType().getType()+"();";
            list.add(str);
        }
        else
        {
            LinkedList list = (LinkedList) variableTable.get(currentVariable);
            str = "      $this->local[\""+node.getIdentifier().getText()+"\"];\n";
            list.add(str);
        }
    }

    public void inAManyIdentifiers(AManyIdentifiers node)
    {

        Node tmpnode=node;
        /*keep going until you hit a variable;*/
        while(tmpnode instanceof PIdentifiers)
        {
            tmpnode = tmpnode.parent();
        }

        /*get rid of case keepmany and removemany from tuples*/
        if(!(tmpnode instanceof AVariable))
        {
            return;
        }

        LinkedList list = (LinkedList) variableTable.get(currentVariable);
        String str = "       $this->local[\""+node.getIdentifier().getText()+"\"]; \n";
        list.add(str);
    }

    public void inAVariable(AVariable node)
    {
        currentVariable = node;
        variableTable.put(node, new LinkedList());
    }

    public void outAVariable(AVariable node)
    {
        Integer cc = new Integer(ccCounter);
        ccCounter++;
        String str = "    if(!$this->cc["+cc.toString()+"]) {\n";
        Iterator itera = ((LinkedList)variableTable.get(node)).listIterator();
        while(itera.hasNext())
        {
            str += (String) itera.next();
        }
        str += "      $this->cc["+cc.toString()+"] = true; \n    }\n ";
        codeTable.put(node,str);
    }

    /***** compoundstm   *****/

    public void outACompoundstm(ACompoundstm node)
    {
        Iterator stms = node.getStm().listIterator();
        Iterator vars = node.getVariable().listIterator();
        String str = "";
        // generate the variables
        while(vars.hasNext())
        {
            str+=((String)codeTable.get(vars.next()));
        }
        PStm pnode;
        Iterator calls;
        // generate the code 
        while(stms.hasNext())
        {
            /*place call pushes*/
            pnode = (PStm) stms.next();
            if(pushcallTable.containsKey(pnode))
            {
                calls = ((LinkedList)pushcallTable.get(pnode)).listIterator();
                while(calls.hasNext())
                {
                    str+= (String) calls.next();
                }
            }
            /*the acutal statement*/
            str+=(String)codeTable.get(pnode);
        }
        codeTable.put(node,str);
    }

}
