/*
    Main Class of Wig compiler.
 */
import java.util.*;
import java.io.*;

import wig.node.*;
import wig.compiler.*;
import wig.analysis.*;
import wig.parser.*;
import wig.lexer.*;
import wig.weeder.*;
import wig.symbol.*;
import wig.type.*;
import wig.code.*;

import java.io.File;
import java.io.FileReader;
import java.io.PushbackReader;

class Compiler
{
    private static String outputDir="."+File.separator;

    public static void main(String[] argv)
    {
        Start tree = null;
        Parser p;
        Iterator filesIter;
        String srcfile,outfile;
        LinkedList theProgram = new LinkedList();
        filesIter = processCmdLine(argv).iterator();

        /*iterator on list of input classfiles*/
        while(filesIter.hasNext())
        {
            srcfile = (String)filesIter.next();
            outfile = srcfile.replaceFirst(".wig$",".php");
            outfile = outputDir+outfile;

            File outfileObject = new File(outfile);
            outfileObject = outfileObject.getParentFile();
            if(!outfileObject.exists())
            {
                System.out.println(outfileObject);
                outfileObject.mkdirs();
            }
            try
            {
                p = new Parser (new Lexer(new PushbackReader(
                                              new FileReader(srcfile),5000)));
                tree = p.parse();

                if(PrintError.hasError())
                {
                    System.out.print("In Scanning and Prasing: it has ");
                    PrintError.printErrorsCount(); //print number of errors
                    System.exit(1);
                }

                //tree.apply(new Weeder()); /*also add import wig.weeder.*;*/
                AnalysisDataManager manager = new AnalysisDataManager();

                tree.apply(new FirstSymWalk(manager));
                if(PrintError.hasError())
                {
                    System.out.print("In First Pass of Symbol Checking: it has ");
                    PrintError.printErrorsCount(); //print number of errors
                    System.exit(1);
                }
                
                tree.apply(new SecondSymWalk(manager));
                if(PrintError.hasError())
                {
                    System.out.print("In Second Pass of Symbol Checking: it has ");
                    PrintError.printErrorsCount(); //print number of errors
                    System.exit(1);
                }
                
                tree.apply(new TypeWalk(manager));
                if(PrintError.hasError())
                {
                    System.out.print("In Type Checking: it has ");
                    PrintError.printErrorsCount(); //print number of errors
                    System.exit(1);
                }
                
                tree.apply(new PreCodeWalk(manager));
                if(PrintError.hasError())
                {
                    System.out.print("In PreCodeWalk: it has ");
                    PrintError.printErrorsCount(); //print number of errors
                    System.exit(1);
                }

                PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(outfile)));
                tree.apply(new CodeWalk(manager, out));
                out.close();
                if(PrintError.hasError())
                {
                    System.out.print("In Code Generating: it has ");
                    PrintError.printErrorsCount(); //print number of errors
                    System.exit(1);
                }

                System.out.println("Finished Compilation of "+outfile);

            }
            catch(LexerException le)
            {

                System.out.println("Lexer : "+le.getMessage());
                //le.printStackTrace();
            }
            catch(ParserException pe)
            {
                System.out.println("Parser : "+pe.getMessage());
                //pe.printStackTrace();
            }
            catch(Exception e)
            {
                System.out.println("Other Exception : "+e.getMessage());
                e.printStackTrace();
            }

        }
    }//end of main

    /**
     * Processes options and extracts the actual arguments (filenames).
     */
    private static List processCmdLine(String[] args)
    {
        String arg;
        List files = new LinkedList();

        if(args.length != 0)
        {
            for (int i = 0; i < args.length; i++)
            {
                arg = args[i].intern(); // optimization: can use '==' for .equals
                if ((arg == "-h") || (arg == "--help"))
                {
                    System.out.println("Usage: wig2php [OPTION]... [FILE]...");
                    System.out.println("Options:");
                    System.out.println("  -d dir    specify the output directory");
                    System.out.println();
                    System.out.println("Input files should have extension '.wig' for regular classes");
                    System.exit(0);
                }
                // We add a output directory option
                else if (arg == "-d")
                {
                    i++;
                    if(arg.endsWith("\\") || arg.endsWith("/"))
                        outputDir=args[i];
                    else
                        outputDir=args[i]+File.separator;
                }
                else if (arg.startsWith("-"))
                {
                    System.err.println("Unrecognized option: "+arg);
                    System.exit(-1);
                }
                else
                {
                    if(arg.startsWith(".\\") || arg.startsWith("./"))
                        arg = arg.substring(2);
                    files.add(arg);
                }
            }
        }
        else
        {
            System.out.println("Usage: wig2php [OPTION]... [FILE]...");
            System.out.println("Options:");
            System.out.println("  -d dir    specify the output directory");
            System.out.println();
            System.out.println("Input files should have extension '.wig' for regular classes");
            System.exit(0);
        }

        return files;
    }

}

