/*
    This Class print out the error in compilation and stop the compiler.
 */
package wig.compiler;

public class PrintError
{
    private static int errorsCount = 0;

    public static boolean hasError()
    {
        return errorsCount>0;
    }

    /*********************************
     * Errors                        *
     * -outputs the error            *
     *  and goes on with compilation *
     *********************************/
    public static void error(String msg, int lineno)
    {
        errorsCount++;
        System.out.println(lineno+": "+msg);
    }

    public static void error(String file, String msg, int lineno, int pos)
    {
        errorsCount++;
        System.out.println(file+":"+lineno+","+pos+": "+msg);
    }

    public static void error(int line)
    {
        errorsCount++;
        System.out.println(line + ": Confused by earlier errors");
        System.out.println(errorsCount + " errors");
        System.exit(1);
    }


    /************************************
     * Fatal Errors                     *
     * -outputs the error and quits the *
     *  compilation                     *
     ************************************/

    public static void fatalError(String file, String phase,
                                  String msg, int lineno)
    {
        System.out.println("Fatal error in "+phase+
                           " at line "+lineno+": "+msg);
        System.out.println(
            "This should never happen! Please submit a bug report.");
        System.exit(1);
    }
    public static void fatalError(String phase, String msg)
    {
        System.out.println("Fatal error in "+phase+": "+msg);
        System.out.println(
            "This should never happen! Please submit a bug report.");
        System.exit(1);
    }

    public static void fatalError(String msg, int lineno)
    {
        errorsCount++;
        System.out.println(lineno+": "+msg);
        System.out.println(errorsCount + " errors");
        System.exit(1);
    }

    /*
      Method that outputs the number of errors
    */
    public static void printErrorsCount()
    {
        if (errorsCount > 0)
        {
            System.out.println(errorsCount+" error"+(errorsCount>1 ? "s" : ""));
            System.exit(1);
        }
    }
}





