/*
    Create the Symbol and put into symbol table for:
    HtmlConst and hole
    schema and their fields
    function
    session
 */
package wig.symbol;

import java.util.*;
import wig.analysis.*;
import wig.node.*;
import wig.symbol.*;
import wig.compiler.*;

public class FirstSymWalk extends DepthFirstAdapter
{
    private SymbolTable curScope;
    private Hashtable cstToScopeMap;

    public FirstSymWalk(AnalysisDataManager manager)
    {
        curScope = manager.getGlobalTable();
        cstToScopeMap = manager.getCstToScopeMap();
    }

    /**** html symbol table *****/
    public void inAHtml(AHtml node)
    {
        String id = node.getIdentifier().getText();
        if(curScope == null)
            PrintError.fatalError("Creating SymbolTable","Global Table is NULL.");
        if(curScope.defSymbol(id))
        {
            PrintError.error("Error: Redefinition of html identifier"+
                             id,node.getIdentifier().getLine());
        }
        else
        {
            curScope.putSymbol(id,Const.html,node);
            //scope in
            curScope = curScope.scopeSymbolTable(); /*for holes*/
            cstToScopeMap.put(node,curScope);
        }
    }

    public void inAHoleHtmlbody(AHoleHtmlbody node)
    {
        String id = node.getIdentifier().getText();
        if(curScope == null)
            return;
        if(!curScope.defSymbol(id))
        {
            curScope.putSymbol(id,Const.hole,node);
        } /*else {allow this, so you can re-use variables}*/
    }

    public void outAHtml(AHtml node)
    {
        //scope out
        curScope = curScope.unscopeSymbolTable();
    }



    /**** schema ****/
    public void inASchema(ASchema node)
    {
        String id = node.getIdentifier().getText();
        if(curScope.defSymbol(id))
        {
            PrintError.error("Error: Redefinition of schema identifier"+
                             id,node.getIdentifier().getLine());
        }
        else
        {
            curScope.putSymbol(id,Const.schema,node);
            //scope in
            curScope = curScope.scopeSymbolTable();
            cstToScopeMap.put(node,curScope);
        }
    }

    public void inAField(AField node)
    {
        String id = node.getIdentifier().getText();
        if(curScope.defSymbol(id))
        {
            PrintError.error("Error: Redefinition of field identifier in schema"+
                             id,node.getIdentifier().getLine());
        }
        else
        {
            // put a new field in the Symbol table of schema
            if(node.getSimpletype() instanceof AIntSimpletype)
                curScope.putSymbol(id, Const.field, node, Const.tInt);
            else if(node.getSimpletype() instanceof ABoolSimpletype)
                curScope.putSymbol(id, Const.field, node, Const.tBool);
            else if(node.getSimpletype() instanceof AStringSimpletype)
                curScope.putSymbol(id, Const.field, node, Const.tString);
            else if(node.getSimpletype() instanceof AVoidSimpletype)
                curScope.putSymbol(id, Const.field, node, Const.tVoid);
        }
    }

    public void outASchema(ASchema node)
    {
        curScope = curScope.unscopeSymbolTable();
    }



    /**** functions ****/
    public void inAFunction(AFunction node)
    {
        String id = node.getIdentifier().getText();
        //check the redefinition
        if(curScope.defSymbol(id))
        {
            PrintError.error("Error: Redefinition of function identifier"+
                             id,node.getIdentifier().getLine());
        }
        // check the return type of function
        else if(node.getType() instanceof ATupleType)
        {
            // Tuple return type
            TIdentifier schemaid = ((ATupleType)node.getType()).getIdentifier();
            Symbol s = curScope.getSymbol(schemaid.getText());
            if(s==null)
            {
                PrintError.error("Error: "+" the schema identifier " +
                                 schemaid.getText() + " is not defined",
                                 node.getIdentifier().getLine());
            }
            else
            {
                curScope.putSymbol(id,Const.function,node,new Type(schemaid.getText()));
            }
        }
        else if(node.getType() instanceof ASimpleType)
        {
            //simple Tuple Type
            if(((ASimpleType)node.getType()).getSimpletype() instanceof AIntSimpletype)
                curScope.putSymbol(id, Const.function, node, Const.tInt);
            else if(((ASimpleType)node.getType()).getSimpletype() instanceof ABoolSimpletype)
                curScope.putSymbol(id, Const.function, node, Const.tBool);
            else if(((ASimpleType)node.getType()).getSimpletype() instanceof AStringSimpletype)
                curScope.putSymbol(id, Const.function, node, Const.tString);
            else if(((ASimpleType)node.getType()).getSimpletype() instanceof AVoidSimpletype)
                curScope.putSymbol(id, Const.function, node, Const.tVoid);
        }
    }



    /**** session ****/
    public void inASession(ASession node)
    {
        String id = node.getIdentifier().getText();
        if(curScope.defSymbol(id))
        {
            PrintError.error("Error: Redefinition of session identifier"+
                             id,node.getIdentifier().getLine());
        }
        else
        {
            curScope.putSymbol(id, Const.session, node);
        }
    }


}


