/*
    In this Walk:
    define all the symbol when it is declare in the program.
    check the symbol is defined in this scope or parent scope when use. 
 */
package wig.symbol;

import java.util.*;
import wig.analysis.*;
import wig.node.*;
import wig.symbol.*;
import wig.compiler.*;

public class SecondSymWalk extends DepthFirstAdapter
{
    private SymbolTable curScope;
    private Hashtable cstToScopeMap;
    private SymbolTable globaltable;

    public SecondSymWalk(AnalysisDataManager manager)
    {
        curScope = manager.getGlobalTable();
        globaltable = curScope;
        cstToScopeMap = manager.getCstToScopeMap();
    }

    // pass the name of a schema or a html, get the symbol table that hold
    private SymbolTable getSymbolTableOf(String name)
    {
        //get the Node defining the schema
        Node node = curScope.getSymbol(name).getNode();
        //get the Symbol Table that define the field inside the schema
        return (SymbolTable) cstToScopeMap.get(node);
    }

    /*
     * identifiers =
     *       {one}  identifier |
     *       {many} identifiers comma identifier;
     * global and local declarations are dealt with here!
     */
    public void inAOneIdentifiers(AOneIdentifiers node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();

        Node tmpnode=node;
        /*keep going until you hit a variable;*/
        while(tmpnode instanceof PIdentifiers)
        {
            tmpnode = tmpnode.parent();
        }

        /*get rid of case keepmany and removemany from tuples*/
        if(!(tmpnode instanceof AVariable))
        {
            return;
        }
        AVariable variable = (AVariable) tmpnode;

        if(curScope.defSymbol(name))
        {
            PrintError.error("Error: "+" identifier " + id.getText()
                             + " is already defined " , id.getLine());
        }
        else if(variable.getType() instanceof ATupleType)
        {
            //tuple: get the name of the schema and get the symbol of that 
            TIdentifier schemaid = ((ATupleType)variable.getType()).getIdentifier();
            Symbol s = curScope.getSymbol(schemaid.getText());
            // if null means, it haven't been defined
            if(s==null)
            {
                PrintError.error("Error: "+" the schema identifier " +
                                 schemaid.getText() + " is not defined",id.getLine());
            }
            else
            {
                SymbolTable schemaSyms = getSymbolTableOf(schemaid.getText());
                curScope.putSymbol(name,Const.variable,node,new Type(schemaid.getText(), schemaSyms));
            }
        }
        else if(variable.getType() instanceof ASimpleType)
        {
            // simple variable
            if(((ASimpleType)variable.getType()).getSimpletype() instanceof AIntSimpletype)
            {
                curScope.putSymbol(name, Const.variable, node, Const.tInt);
            }
            else if(((ASimpleType)variable.getType()).getSimpletype() instanceof ABoolSimpletype)
            {
                curScope.putSymbol(name, Const.variable, node, Const.tBool);
            }
            else if(((ASimpleType)variable.getType()).getSimpletype() instanceof AStringSimpletype)
            {
                curScope.putSymbol(name, Const.variable, node, Const.tString);
            }
            else if(((ASimpleType)variable.getType()).getSimpletype() instanceof AVoidSimpletype)
            {
                curScope.putSymbol(name, Const.variable, node, Const.tVoid);
            }
        }
        cstToScopeMap.put(node, curScope);
    }


    public void inAManyIdentifiers(AManyIdentifiers node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();

        Node tmpnode=node;
        /*keep going until you hit a variable;*/
        while(tmpnode instanceof PIdentifiers)
        {
            tmpnode = tmpnode.parent();
        }
        /*
            Prevent those:
                {keep_many}   tuple_exp keep l_par identifiers r_par |
                {remove_many} tuple_exp remove l_par identifiers r_par |
         */
        if(!(tmpnode instanceof AVariable))
            return;

        AVariable variable = (AVariable) tmpnode;

        if(curScope.defSymbol(name))
        {
            PrintError.error("Error: "+" identifier " +
                             id.getText() + " is already defined ",id.getLine());
        }
        else if(variable.getType() instanceof ATupleType)
        {
            TIdentifier schemaid = ((ATupleType)variable.getType()).getIdentifier();
            Symbol s = curScope.getSymbol(schemaid.getText());
            if(s==null)
            {
                PrintError.error("Error: "+" the schema identifier " +
                                 schemaid.getText() + " is not defined ", id.getLine());
            }
            else
            {
                SymbolTable schemaSyms = getSymbolTableOf(schemaid.getText());
                curScope.putSymbol(name,Const.variable,node,new Type(schemaid.getText(), schemaSyms));
            }
        }
        else if(variable.getType() instanceof ASimpleType)
        {
            if(((ASimpleType)variable.getType()).getSimpletype() instanceof AIntSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tInt);
            }
            else if(((ASimpleType)variable.getType()).getSimpletype() instanceof ABoolSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tBool);
            }
            else if(((ASimpleType)variable.getType()).getSimpletype() instanceof AStringSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tString);
            }
            else if(((ASimpleType)variable.getType()).getSimpletype() instanceof AVoidSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tVoid);
            }
            cstToScopeMap.put(node, curScope);
        }
    }

    /*function arguments*/
    public void inAFunction(AFunction node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();
        if(curScope.defSymbol(name))
        {
            // scope in
            curScope = curScope.scopeSymbolTable();
            cstToScopeMap.put(node,curScope);
        }
        else
        {
            PrintError.error("This should not happen because the function "+
                             "were define in the prior Symbol Table scope. "+
                             "Function: "+id.getText() ,id.getLine());
        }
    }

    public void inAArgument(AArgument node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();
        if(curScope.defSymbol(name))
        {
            PrintError.error("Error: "+" identifier " + id.getText()
                             + " is already defined ", id.getLine());
        }
        // similar to define a variable (see inAManyIdentifiers)
        else if(node.getType() instanceof ATupleType)
        {
            TIdentifier schemaid = ((ATupleType)node.getType()).getIdentifier();
            Symbol s = curScope.getSymbol(schemaid.getText());
            if(s==null)
            {
                PrintError.error("Error: "+" the schema identifier " +
                                 schemaid.getText() + " is not defined ", id.getLine());
            }
            else
            {
                SymbolTable schemaSyms = getSymbolTableOf(schemaid.getText());
                curScope.putSymbol(name,Const.variable,node,new Type(schemaid.getText(), schemaSyms));
            }
        }
        else if(node.getType() instanceof ASimpleType)
        {
            if(((ASimpleType)node.getType()).getSimpletype() instanceof AIntSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tInt);
            }
            else if(((ASimpleType)node.getType()).getSimpletype() instanceof ABoolSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tBool);
            }
            else if(((ASimpleType)node.getType()).getSimpletype() instanceof AStringSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tString);
            }
            else if(((ASimpleType)node.getType()).getSimpletype() instanceof AVoidSimpletype)
            {
                curScope.putSymbol(name,Const.variable,node,Const.tVoid);
            }
        }
        cstToScopeMap.put(node, curScope);
    }

    public void outAFunction(AFunction node)
    {
        //scope out
        curScope = curScope.unscopeSymbolTable();
    }


    /*session inwards*/
    public void inASession(ASession node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();
        if(curScope.defSymbol(name))
        {
            //scope in
            curScope = curScope.scopeSymbolTable();
            cstToScopeMap.put(node,curScope);
        }
        else
        {
            PrintError.error("This should not happen because this session "+
                             "was defined in the prior Symbol Table scope. "+
                             "Session: "+id.getText(), id.getLine());
        }
    }

    public void outASession(ASession node)
    {
        //scope out
        curScope = curScope.unscopeSymbolTable();
    }


    /*compoundstm*/
    public void inACompoundstm(ACompoundstm node)
    {
        /*don't scope in sessions or functions because it was already done*/
        if(!((Node) node.parent() instanceof PFunction) &&
                !((Node) node.parent() instanceof PSession))
        {
            //scope in
            curScope = curScope.scopeSymbolTable();
            cstToScopeMap.put(node,curScope);
        }

    }

    public void outACompoundstm(ACompoundstm node)
    {
        if(!((Node) node.parent() instanceof PFunction) &&
                !((Node) node.parent() instanceof PSession))
        {
            //scope out
            curScope = curScope.unscopeSymbolTable();
        }
    }


    /**** output in plugs and exit (treat as defining new variables in side a function) ****/
    public void inAPlugDocument(APlugDocument node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();
        if(curScope.getSymbol(name) != null)
        {
            curScope = curScope.scopeSymbolTable();
            cstToScopeMap.put(node,curScope);
        }
        else
        {
            PrintError.error("Error: "+" the html identifier " +
                             name + " was not defined",id.getLine());
        }
    }

    public void inAIDDocument(APlugDocument node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();
        if(curScope.getSymbol(name) == null)
        {
            PrintError.error("Error: "+" the html identifier " +
                             name + " was not defined",id.getLine());
        }
    }

    public void inAPlug(APlug node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();
        if(curScope.defSymbol(name))
        {
            PrintError.error("Error: "+" identifier " + id.getText()
                             + " is already defined " , id.getLine());
        }
        else
        {

            // report a "warning" if the user want to plug a undefined hole in html

            // keep going until hit a document;
            Node tmpnode = node;
            while(!(tmpnode instanceof APlugDocument))
            {
                tmpnode = tmpnode.parent();
            }

            TIdentifier htmlid = ((APlugDocument)tmpnode).getIdentifier();
            Symbol h = curScope.getSymbol(htmlid.getText());

            if(h!=null)
            {
                //Get the Symbol Table in the html and finde the Symbol inside that
                SymbolTable table= ((SymbolTable) cstToScopeMap.get( h.getNode() ) );
                Symbol hole = table.getSymbol(name);

                /*
                 * hole is not defined
                 */
                if(hole==null)
                {
                    PrintError.error("Warning: "+" the hole identifier " +
                                     hole + " was not defined",id.getLine());
                }
                else
                {
                    curScope.putSymbol(name, Const.hole, node);
                }
            }//else not need to print out error, it have been done in inAPlugDocument()
        }

        cstToScopeMap.put(node, curScope); //in fact it doesn't really needed
    }

    public void outAPlugDocument(APlugDocument node)
    {
        curScope = curScope.unscopeSymbolTable();
    }




    /**** inputs in receives (treat as defining new variables in side a compound stm) ****/
    public void inAReceive(AReceive node)
    {
        curScope = curScope.scopeSymbolTable();
        cstToScopeMap.put(node,curScope);
    }

    public void inAInput(AInput node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();
        if(curScope.defSymbol(name))
        {
            PrintError.error("Error: "+" identifier " + id.getText()
                             + " is already defined " , id.getLine());
        }   //in fact reuse of the input may be allowed, but I don't think that is really useful
        else
        {
            curScope.putSymbol(name, Const.hole, node);
        }
        cstToScopeMap.put(node, curScope); //in fact it doesn't really needed
    }

    public void outAReceive(AReceive node)
    {
        curScope = curScope.unscopeSymbolTable();
    }


    /**** function call in side function or session ****/
    public void inACallBaseExp(ACallBaseExp node)
    {
        TIdentifier id = node.getIdentifier();
        String functionName = id.getText();

        Symbol funcSymbol = curScope.getSymbol(functionName);
        if(funcSymbol!=null)
        {
            /*
                only check the number of arguemnt in here
                because the type will be checked in type checker
             */

            //check the argument of the function match
            PExps temnode1 = node.getExps();

            //find the number of the argument
            int arguCount1=1;
            while(temnode1 instanceof AManyExps)
            {
                temnode1 = ((AManyExps)temnode1).getExps();
                arguCount1++;
            }

            //get the Arguments list in the defination of the function
            PArguments temnode2 = ((AFunction)funcSymbol.getNode()).getArguments();
            //find the number of the argument
            int arguCount2=1;
            while(temnode2 instanceof AManyArguments)
            {
                temnode2 = ((AManyArguments)temnode2).getArguments();
                arguCount2++;
            }

            if(arguCount1 > arguCount2)
            {
                PrintError.error("Error: "+" too many arguments of " + functionName , id.getLine());
            }
            else if(arguCount1 < arguCount2)
            {
                PrintError.error("Error: "+" not enough arguments of " + functionName , id.getLine());
            }
        }
        else
        {
            PrintError.error("Error: "+" function " + functionName
                             + " was not defined " , id.getLine());
        }
    }


    /**** Use tuple and variables ****/

    // check the field in tuple is defined
    public void inAQualifiedLvalue(AQualifiedLvalue node)
    {
        TIdentifier tupleId = node.getLeft();
        String tupleName = tupleId.getText();
        TIdentifier fieldId = node.getRight();
        String fieldName = fieldId.getText();

        Symbol tupleSym = curScope.getSymbol(tupleName);
        if(tupleSym==null)
        {
            PrintError.error("Error: tuple "+ tupleName + " was not defined " , tupleId.getLine());
        }
        else
        {
            if(tupleSym.getType().getField(fieldName)==null)
            {
                PrintError.error("Error: field "+ fieldName + " in " + tupleName
                                 + " was not defined " , tupleId.getLine());
            }
        }
    }


    /**** Using tuple operations of checking the defination in tuple operation ****/
    /*
    	we allows the following tuple operations: (x,y,z is tuple)
           
    	x = y;      (x and y need to have same fields)
    	x = y << z;   (x and y joined with z need to have same fields)
    	x = tuple{field_i=100, field_j="Hi"};   (all field must be init)
    	x = y\+ field_i;            (the field of both sides must be equal)
    	x = y\+(field_i, field_j);  (the field of both sides must be equal)
    	x = y\- field_i;            (the field of both sides must be equal)
    	x = y\-(field_i, field_j);  (the field of both sides must be equal)
           
    	if we find that x is tuple. Then, the whole experssion will treated as a tuple experssion.
    	All the lvalue appear must be tuple. 
    */

    // It is a request to inAOneIdentifiers and inAManyIdentifiers
    private boolean tupleExp = false;
    private boolean rightTupleExp = false;
    private Hashtable leftSymTable=null;
    private Hashtable rightSymTable=null;

    // check the assign exp is a tuple experssion or not
    public void inAAssignExp(AAssignExp node)
    {
        if(node.getLvalue() instanceof ASimpleLvalue)
        {
            TIdentifier id = ((ASimpleLvalue)node.getLvalue()).getIdentifier();
            String name = id.getText();

            Symbol varSym = curScope.getSymbol(name);
            if(varSym!=null)
            {
                //get the name of the schema
                String varType = varSym.getType().getType();
                Type type = new Type(varType);
                if(!(type.equals(Const.tInt) || type.equals(Const.tBool)
                        || type.equals(Const.tVoid) || type.equals(Const.tString)))
                {
                    String schemaName = varType;
                    //get the Node defining the schema

                    Node schemaNode = curScope.getSymbol(schemaName).getNode();
                    //get the Symbol Table that define the field inside the schema
                    SymbolTable schemaTable = (SymbolTable) cstToScopeMap.get(schemaNode);

                    // if it is tuple, enter the tuple experssion mode
                    tupleExp = true;

                    //clone the hash table
                    leftSymTable = (Hashtable)( schemaTable.getTable() ).clone();
                }
            }
        }
    }

    // check the assign exp is a tuple experssion or not
    public void caseTAssign(TAssign node)
    {
        if(tupleExp)
            rightTupleExp=true;
    }

    // due with x = tuple{field_i=100, field_j="Hi"};   (all field must be init)
    public void inAFieldvalue(AFieldvalue node)
    {
        if(tupleExp && leftSymTable != null)
        {
            String initField = node.getIdentifier().getText();
            // init the rightSymTable
            if(rightSymTable == null)
                rightSymTable = new Hashtable(SymbolTable.hashSize);
            //put a key inside the rightSymTable, so that when outAAssignExp
            // rightSymTable and leftSymTable will be matched
            rightSymTable.put(initField, node.getIdentifier());
        }
        else
            PrintError.error("Error: No tuple associated with this operation ",
                             node.getIdentifier().getLine());
    }

    // check the tuple or variable is defined
    // also due with simple variables
    public void inASimpleLvalue(ASimpleLvalue node)
    {
        TIdentifier id = node.getIdentifier();
        String name = id.getText();

        Symbol varSym = curScope.getSymbol(name);
        if(varSym==null)
        {
            PrintError.error("Error: "+ name + " was not defined " , id.getLine());
        }
        else
        {
            if(tupleExp)
            {
                //get the name of the schema
                String varType = varSym.getType().getType();

                if(varType.equals(Const.tInt) || varType.equals(Const.tBool)
                        || varType.equals(Const.tVoid) || varType.equals(Const.tString))
                {
                    //if in a tupleExp, only tuple is allow
                    PrintError.error("Error: "+ name + " is not a tuple " , id.getLine());
                }
                else
                {
                    //get the Node defining the schema
                    Node schemaNode = curScope.getSymbol(varType).getNode();
                    //get the Symbol Table that define the field inside the schema
                    SymbolTable schemaTable = (SymbolTable) cstToScopeMap.get(schemaNode);

                    // if it is the point of rightTuple Experssion
                    if(rightTupleExp == true)
                    {
                        if(rightSymTable == null)
                        {
                            //clone the hash table
                            rightSymTable = (Hashtable)( schemaTable.getTable() ).clone();
                        }
                        else
                        {
                            // add the symbol of the joined tuple into the rightSymTable too
                            Hashtable joining = (Hashtable)( schemaTable.getTable() ).clone();

                            // when x<<y, replace the entry in y for x
                            joining.putAll(rightSymTable);
                            rightSymTable = joining;
                        }
                    }
                }
            }
            //else for simple experssion in normal experssion
        }
    }


    // check the two pair of tuple hashtables quit a tuple experssion if it is
    public void outAAssignExp(AAssignExp node)
    {
        if(tupleExp)
        {
            //match the two Hashtables. if they are different -> left and right hand side tuple are different
            boolean same = true;

            if(leftSymTable != null && rightSymTable != null)
            {
                // get all the fields in left and right .. then compare them
                Enumeration left=leftSymTable.keys();
                Enumeration right=rightSymTable.keys();
                while(left.hasMoreElements() && right.hasMoreElements()) //same
                {
                    String leftField = (String)left.nextElement();
                    String rightField = (String)right.nextElement();
                    same =  leftField.equals(rightField);
                }
                if(left.hasMoreElements() || right.hasMoreElements())
                    same = false;
            }
            else
            {
                same = false;
            }

            int line = node.getAssign().getLine();
            if(same == false)
                PrintError.error("Error: the fields in left hand side tuple is "+
                                 "not matched with the right hand side " , line);

            rightTupleExp=false;
            tupleExp = false;
            leftSymTable=null;
            rightSymTable=null;
        }
    }



    // check the tuple operation of keep
    public void outAKeepTupleExp(AKeepTupleExp node)
    {
        if(tupleExp)
        {
            if(leftSymTable != null && rightSymTable != null)
            {
                String keepField = node.getIdentifier().getText();
                //only keep the entry
                Object keepSymbol = rightSymTable.get(keepField);
                rightSymTable.clear();
                rightSymTable.put(keepField,keepSymbol);
            }
            else
            {
                PrintError.error("Error: No tuple associated with this keep operation ",
                                 node.getKeep().getLine());
            }
        }
        else
            PrintError.error("Error: tuple must be assigned in "
                             + node, node.getKeep().getLine());
    }


    // check the tuple operation of keep
    public void outARemoveTupleExp(ARemoveTupleExp node)
    {
        if(tupleExp)
        {
            if(leftSymTable != null && rightSymTable != null)
            {
                String removeField = node.getIdentifier().getText();
                //only remove the entry
                rightSymTable.remove(removeField);
            }
            else
            {
                PrintError.error("Error: No tuple associated with this remove operation "+node,
                                 node.getRemove().getLine());
            }
        }
        else
            PrintError.error("Error: tuple must be assigned in "
                             + node, node.getRemove().getLine());
    }



    public void outAKeepManyTupleExp(AKeepManyTupleExp node)
    {
        if(tupleExp)
        {
            if(leftSymTable != null && rightSymTable != null)
            {
                String keepField;
                Hashtable tempTable = new Hashtable(SymbolTable.hashSize);
                Node temnode = node.getIdentifiers();
                while(temnode instanceof AManyIdentifiers)
                {
                    keepField  = ((AManyIdentifiers)temnode).getIdentifier().getText(); //without s
                    temnode = ((AManyIdentifiers)temnode).getIdentifiers(); // have s

                    //put the keep fields into the temp hashtable first
                    tempTable.put(keepField, rightSymTable.get(keepField));
                }
                keepField = ((AOneIdentifiers)temnode).getIdentifier().getText();
                tempTable.put(keepField, rightSymTable.get(keepField));

                rightSymTable = tempTable;
            }
            else
            {
                PrintError.error("Error: No tuple associated with this keep operation ",
                                 node.getKeep().getLine());
            }
        }
        else
            PrintError.error("Error: tuple must be assigned in "
                             + node, node.getKeep().getLine());
    }


    public void outARemoveManyTupleExp(ARemoveManyTupleExp node)
    {
        if(tupleExp)
        {
            if(leftSymTable != null && rightSymTable != null)
            {
                String removeField;
                Node temnode = node.getIdentifiers();
                while(temnode instanceof AManyIdentifiers)
                {
                    removeField  = ((AManyIdentifiers)temnode).getIdentifier().getText(); //without s
                    temnode = ((AManyIdentifiers)temnode).getIdentifiers(); // have s

                    // remove all the field
                    rightSymTable.remove(removeField);
                }
                removeField = ((AOneIdentifiers)temnode).getIdentifier().getText();
                rightSymTable.remove(removeField);
            }
            else
            {
                PrintError.error("Error: No tuple associated with this keep operation ",
                                 node.getRemove().getLine());
            }
        }
        else
            PrintError.error("Error: tuple must be assigned in "
                             + node, node.getRemove().getLine());
    }
}

