/*
    SymbolTable contains all the Symbol in this current scope.
    It is also responsible for scope in and out in function or session.
    It can be used to seek symbol and check symbol is defined in this scope.
 */
package wig.symbol;

import java.util.*;
import java.util.Hashtable.*;
import java.util.Iterator.*;
import wig.node.*;

public class SymbolTable
{
    public static final int hashSize = 317; /*prime number*/
    private Hashtable table;
    public SymbolTable next;

    public SymbolTable()
    {
        table = new Hashtable(hashSize);
        next = null;
    }

    public int size()
    {
        return table.size();
    }

    /*return stack with new table ontop*/
    public SymbolTable scopeSymbolTable()
    {
        SymbolTable newtable;
        newtable = new SymbolTable();
        newtable.next = this;
        return newtable;
    }

    /*return stack with popped table*/
    public SymbolTable unscopeSymbolTable()
    {
        return this.next;
    }

    /**
     * create symbol in table if it doesn't already exist
     * then return the symbol
     */
    public Symbol putSymbol(String i, Kind k, Node n)
    {
        Symbol s = (Symbol)table.get(i);
        if(s==null)
        {
            s = new Symbol(i,k,n);
            table.put(i,s);
        }
        return s;
    }
    public Symbol putSymbol(String i, Kind k, Node n, Type t)
    {
        Symbol s = (Symbol)table.get(i);
        if(s==null)
        {
            s = new Symbol(i,k,n,t);
            table.put(i,s);
        }
        return s;
    }

    /**
     * return true if the identifier is defined on
     * the top of the stack
     */
    public boolean defSymbol(String i)
    {
        return (table.get(i)!=null);
    }

    /** return true if the identifier is defined
        somewhere in the stack of symboltables
        defsSymbol means Defined Somewhere Symbol
    **/
    public boolean defsSymbol(String i)
    {
        if(getSymbol(i)==null)
            return false;
        return true;
    }

    /**
     * seek the closely nested symbol if there is one,
     * otherwise null is returned 
     */
    public Symbol getSymbol(String i)
    {
        Symbol s = (Symbol) table.get(i);
        if(s==null&&next!=null)
        {
            s = next.getSymbol(i);
        }
        return s;
    }


    /*does the same thing as getSymbol ... used as a wrapper class*/
    public Symbol seekSymbol(String i)
    {
        return getSymbol(i);
    }

    public Hashtable  getTable()
    {
        return table;
    }

    public String toString()
    {
        Iterator iterator = table.values().iterator();
        Symbol s;
        String str;
        str = "\n---------------------------\n";
        str += "****(name,kind,type)****\n";
        while(iterator.hasNext())
        {
            s = (Symbol) iterator.next();
            str += s.toString() + "\n";
        }
        str += "---------------------------\n";
        return str;
    }

}
