/*
    the Type of the node. (eg. int, bool)
    
    For Tuple type, 
        fieldsTable will contains all the fields name and the type of the fields.
        type will contains the name of the schema identifier.
 */
package wig.symbol;

import java.util.*;

public class Type
{
    private String type;

    //this field table is for storing the field name and type (only for tuple)
    private Hashtable fieldsTable=null;

    public Type(String t)
    {
        type = t;

        //init for tuple
        fieldsTable=new Hashtable();
    }

    //tuple init the fieldsTable
    public Type(String t, SymbolTable schema)
    {
        type = t;

        //init for tuple
        fieldsTable=new Hashtable();

        if(schema != null)
        {
            Iterator iterator = schema.getTable().values().iterator();
            Symbol s;
            String str;
            while(iterator.hasNext())
            {
                s = (Symbol) iterator.next();
                addField(s.getIdentifier(), s.getType());
            }
        }
    }

    //tuple init the fieldsTable
    public Type(String t, Hashtable fields)
    {
        type = t;

        setFieldsTable(fields);
    }

    public String getType()
    {
        return type;
    }
    
    public String toString()
    {
        //if it is tuple, print out the fields two
        if(this.equals(Const.tTuple))
            return this.getType()+this.getFieldsTable().toString();
        else
            return this.getType();
    }

    public boolean equals(String i)
    {
        return i.equals(getType());
    }

    public boolean equals(Type i)
    {
        if(
            type.equals(Const.tInt.getType()) ||
            i.getType().equals(Const.tInt.getType())||
            type.equals(Const.tBool.getType()) ||
            i.getType().equals(Const.tBool.getType())||
            type.equals(Const.tVoid.getType()) ||
            i.getType().equals(Const.tVoid.getType())||
            type.equals(Const.tString.getType()) ||
            i.getType().equals(Const.tString.getType())
        )
            //if one of that is permittive type
        {
            return i.getType().equals(getType());
        }
        else if(type.equals(Const.tTuple.getType()) ||
                i.getType().equals(Const.tTuple.getType()))
            //if just want to know where it is tuple or not
        {
            return true;
        }
        else
            //if it is tuple -> compare the fields type
        {
            return fieldsTable.equals(i.getFieldsTable());
        }
    }

    public Hashtable getFieldsTable()
    {
        return fieldsTable;
    }

    public Hashtable cloneFieldsTable()
    {
        return (Hashtable)fieldsTable.clone();
    }

    public void setFieldsTable(Hashtable newFieldsTable)
    {
        fieldsTable = newFieldsTable;
    }

    public Type getField(String fieldName)
    {
        if(fieldsTable.containsKey(fieldName))
            return (Type)fieldsTable.get(fieldName);
        return null;
    }

    public void addField(String fieldName, Type type)
    {
        fieldsTable.put(fieldName, type);
    }

    public Object removeField(String fieldName)
    {
        return fieldsTable.remove(fieldName);
    }

    public void clearFields()
    {
        fieldsTable.clear();
    }
}
